import numpy as np
from numpy.linalg import inv
from scipy.stats import invgamma
from scipy.stats import multivariate_normal as mvn

from pssm.filters import KalmanFilter


class FFBS:
    def __init__(self, structure, m0, C0, vprior, wprior, data):
        self._structure = structure
        self._m0 = m0
        self._C0 = C0
        self._vprior = vprior
        self._wprior = wprior
        self._data = data
        self._nobs = len(data)

    @staticmethod
    def backward_sampling_step(theta, m, C, W):
        _inv = inv(C + W)
        mean = m + np.dot(C, np.dot(_inv, theta - m).A1).A1
        cov = C - C * np.transpose(C * _inv)

        return mvn(mean=mean, cov=cov).rvs()

    @staticmethod
    def backward_sampling(ms, Cs, W):
        T = len(ms) - 1
        theta = mvn(mean=np.asarray(ms[T]), cov=Cs[T]).rvs()
        thetas = [theta]
        for t in reversed(range(1, T)):
            theta = FFBS.backward_sampling_step(theta, ms[t], Cs[t], W)
            thetas.append(theta)
        return list(reversed(thetas))

    def _states(self, V, W):
        ms = [self._m0]
        Cs = [self._C0]
        structure = self._structure
        structure._W = W
        kf = KalmanFilter(structure=structure, V=V)
        for t in range(1, self._nobs):
            m, C = kf.filter(self._data[t], ms[t - 1], Cs[t - 1])
            ms.append(m)
            Cs.append(C)
        return ms, Cs

    def run(self, V, W, states=False):
        Ft = self._structure.F.T
        n = self._nobs - 1

        ms, Cs = self._states(V=V, W=W)

        thetas = FFBS.backward_sampling(ms, Cs, W)

        ssy = np.zeros(n)
        sstheta = np.zeros((n, len(self._wprior)))
        for t in range(1, n):
            ssy[t] = np.power(self._data[t] - np.dot(Ft, thetas[t]), 2.0)
            sstheta[t] = np.power(
                thetas[t] - np.dot(self._structure.G, thetas[t - 1]), 2.0)

        v_rate = self._vprior + 0.5 * sum(ssy)
        _V = invgamma(self._vprior + 0.5 * n, scale=v_rate).rvs()

        w_rate = [self._vprior + 0.5 * d for d in sstheta.sum(axis=0)]
        _W = np.diag(
            [invgamma(prior + 0.5 * n, scale=scale).rvs()
             for scale, prior in zip(w_rate, self._wprior)])

        if states:
            return _W, _V, thetas
        else:
            return _W, _V



import abc
import numpy as np
from scipy.stats import multivariate_normal as mvn, poisson
import math

from pssm.utils import ilogit
from pssm.structure import MultivariateStructure

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})


class DLM(ABC):

    def __init__(self, structure, state_prior=None):
        self._structure = structure
        self._Ft = np.transpose(structure.F)
        if state_prior is not None:
            self._current_state = state_prior
        else:
            self._current_state = np.zeros(structure.W.shape[0])

    @property
    def structure(self):
        return self._structure

    @property
    def current_state(self):
        return self._current_state

    @abc.abstractmethod
    def _eta(self, _lambda):
        pass

    @abc.abstractmethod
    def _sample_obs(self, mean):
        pass

    def state(self, previous):
        mean = np.squeeze(np.asarray(np.dot(self._structure.G, previous)))
        return mvn(mean=mean, cov=self._structure.W, allow_singular=True).rvs()

    def observation(self, state):
        mean = self._eta(np.dot(self._Ft, state))
        return self._sample_obs(mean)

    def next(self):
        return self.__next__()

    def __next__(self):
        self._current_state = self.state(self._current_state)
        obs = self.observation(self._current_state)
        return self._current_state, obs


class NormalDLM(DLM):
    """
    An instance of a Normal DLM
    """

    def __init__(self, structure, V, state_prior=None):
        super(NormalDLM, self).__init__(structure=structure,
                                        state_prior=state_prior)
        self._V = np.matrix([[V]])

    def _eta(self, _lambda):
        return _lambda

    def _sample_obs(self, mean):
        return mvn(mean=mean, cov=self._V).rvs()


class PoissonDLM(DLM):
    """
    An instance of a Poisson DLM
    """

    def __init__(self, structure, state_prior=None):
        super(PoissonDLM, self).__init__(structure=structure,
                                         state_prior=state_prior)

    def _eta(self, _lambda):
        return math.exp(_lambda)

    def _sample_obs(self, mean):
        return poisson(mean).rvs()


class BinomialDLM(DLM):
    """
    An instance of a Binomial DLM
    """

    def __init__(self, structure, categories=1, state_prior=None):
        super(BinomialDLM, self).__init__(structure=structure,
                                          state_prior=state_prior)
        self._categories = categories

    def _eta(self, _lambda):
        return ilogit(_lambda)

    def _sample_obs(self, mean):
        return np.random.binomial(n=self._categories, p=mean)


class CompositeDLM(DLM):
    """
    A multivariate composite DLM
    """

    def _sample_obs(self, mean):
        pass

    def _eta(self, _lambda):
        pass

    def __init__(self, *dglms):
        super(CompositeDLM, self).__init__(
            MultivariateStructure.build(*[dglm.structure for dglm in dglms]))
        self._dglms = dglms

    def observation(self, state):
        lambdas = np.dot(self._Ft, state)
        ys = []
        for i in range(len(self._dglms)):
            dglm = self._dglms[i]
            eta = dglm._eta(lambdas[i])
            ys.append(dglm._sample_obs(eta))

        return np.array(ys)



import numpy as np
from numpy.linalg import inv


class KalmanFilter:
    def __init__(self, structure, V):
        self._structure = structure
        self._V = V
        self._Ft = np.transpose(self._structure.F)

    def _prior_state(self, m):
        a = np.dot(self._structure.G, m).A1
        return a

    def _prior_covariance(self, C):
        R = self._structure.G * C * self._structure.G.T + self._structure.W
        return R

    def _innovation(self, a, y):
        e = y - np.asscalar(self._Ft.dot(a))
        return e

    def _residual_covariance(self, R):
        Q = self._Ft * R * self._structure.F + self._V
        return Q

    def _gain(self, R, Q):
        K = R * self._structure.F * inv(Q)
        return K

    def filter(self, y, m, C):
        # Predicted (a priori) state estimate
        a = self._prior_state(m)

        # Predicted (a priori) estimate covariance
        R = self._prior_covariance(C)

        # Innovation or measurement residual
        e = self._innovation(a, y)

        # Innovation (or residual) covariance
        Q = self._residual_covariance(R)

        # Kalman gain
        K = self._gain(R, Q)

        new_m = a + K.A1 * e
        new_C = R - K * Q * K.T
        return new_m, new_C



import numpy as np
from scipy.linalg import block_diag


class DimensionError(Exception):
    pass


class UnivariateStructure:
    def __init__(self, F, G, W):
        self._F = F
        self._G = G
        self._W = W

    @property
    def F(self):
        return self._F

    @property
    def G(self):
        return self._G

    @property
    def W(self):
        return self._W

    def __add__(self, other):
        # type: (UnivariateStructure) -> UnivariateStructure
        F = np.vstack((self.F, other.F))
        G = block_diag(*[self.G, other.G])
        W = block_diag(*[self.W, other.W])

        return UnivariateStructure(F=F, G=G, W=W)

    @staticmethod
    def locally_constant(W):
        # type: (float) -> UnivariateStructure
        F = np.matrix([[1]])
        G = np.matrix([[1]])
        W = np.matrix([[W]])
        return UnivariateStructure(F=F, G=G, W=W)

    @staticmethod
    def locally_linear(W):
        # type: (ndarray) -> UnivariateStructure
        F = np.matrix([[1], [0]])
        G = np.matrix([[1, 1], [0, 1]])
        return UnivariateStructure(F=F, G=G, W=W)

    @staticmethod
    def arma(p, betas, W):
        # type: (ndarray, List[float], float) -> UnivariateStructure
        if p < 1:
            raise ValueError("`p` must be 1 or higher.")
        if len(betas) != p:
            raise ValueError("`betas` must have `p` elements.")

        F = np.transpose(np.matrix([[1.0] + [0.0] * (p - 1)]))
        if p == 1:  # can simplify for ARMA(1)
            G = np.identity(p) * betas[0]
        else:
            G = np.identity(p - 1)
            G = np.hstack((G, np.zeros((p - 1, 1))))
            G = np.vstack((betas, G))
        _W = np.zeros((p, p))
        _W[0, 0] = W
        return UnivariateStructure(F=F, G=G, W=_W)

    @staticmethod
    def cyclic_fourier(period, harmonics, W):
        # type: (int, int, ndarray) -> UnivariateStructure
        om = 2.0 * np.pi / period
        harmonic1 = np.identity(2) * np.cos(om)
        harmonic1[0, 1] = np.sin(om)
        harmonic1[1, 0] = -harmonic1[0, 1]

        _G = [None] * harmonics
        _G[0] = np.copy(harmonic1)
        if harmonics > 1:
            for i in np.arange(1, harmonics):
                _G[i] = np.dot(_G[i - 1], harmonic1)
            G = block_diag(*_G)
        else:
            G = harmonic1

        F = np.transpose(np.matrix([[1.0, 0.0] * harmonics]))

        return UnivariateStructure(F=F, G=G, W=W)


class MultivariateStructure:
    def __init__(self, F, G, W):
        self._F = F
        self._G = G
        self._W = W

    @property
    def F(self):
        return self._F

    @property
    def G(self):
        return self._G

    @property
    def W(self):
        return self._W

    @staticmethod
    def build(*dglms):
        _F = block_diag(*[model.F for model in dglms])
        _G = block_diag(*[model.G for model in dglms])
        _W = block_diag(*[model.W for model in dglms])
        return MultivariateStructure(F=_F, G=_G, W=_W)



import numpy as np
from numpy import matrix, array
from scipy.linalg import block_diag
from scipy.stats import multivariate_normal
from scipy.stats import binom


class UnivariateModel:
    def __init__(self):
        pass

    @staticmethod
    def compose(A, B):
        composite = UnivariateModel()
        composite.F = np.hstack((A.F, B.F))
        composite.G = block_diag(*[A.G, B.G])
        composite.m0 = np.hstack((A.m0, B.m0))
        composite.C0 = block_diag(*[A.C0, B.C0])
        composite.W = block_diag(*[A.W, B.W])
        composite.V = A.V + B.V
        return composite


class UnivariateLocallyConstantModel(UnivariateModel):
    def __init__(self, m0=0.0, C0=100.0, W=1.0, V=1.0):
        self.F = matrix([1.0])
        self.G = matrix([1.0])
        self.m0 = array([m0])
        self.C0 = matrix([C0])
        self.W = matrix([W])
        self.V = matrix([V])


class UnivariateFourierModel(UnivariateModel):
    def __init__(self, period, harmonics):
        om = 2.0 * np.pi / period
        harmonic1 = np.identity(2) * np.cos(om)
        harmonic1[0, 1] = np.sin(om)
        harmonic1[1, 0] = -harmonic1[0, 1]

        G = [None] * harmonics
        G[0] = np.copy(harmonic1)
        if harmonics > 1:
            for i in np.arange(1, harmonics):
                G[i] = np.dot(G[i - 1], harmonic1)
            self.G = block_diag(*G)
        else:
            self.G = harmonic1

        self.F = np.matrix([[1.0, 0.0] * harmonics])
        self.m0 = array([0.0] * self.G.shape[0])
        self.C0 = np.diag([100.0] * self.G.shape[0])
        self.W = np.diag([1.0] * self.G.shape[0])
        self.V = matrix([1.0])


def stateGenerator(nobs, model):
    thetas = np.empty((nobs, len(model.m0)))
    theta = multivariate_normal(mean=model.m0, cov=model.C0).rvs()
    for t in np.arange(nobs):
        theta = multivariate_normal(mean=np.dot(model.G, theta),
                                    cov=model.W).rvs()
        thetas[t, :] = theta
    return thetas


class MultivariateModel:
    def __init__(self):
        pass


class MultivariateLocallyConstantModel(MultivariateModel):
    def __init__(self, dimension, m0, C0, W, V):
        self._dimension = dimension
        self.F = np.eye(self._dimension)
        self.G = np.eye(self._dimension)
        self.m0 = m0
        self.C0 = C0
        self.W = W
        self.V = V


def ilogit(alpha):
    return 1.0 / (1.0 + np.exp(-alpha))


def MultivariateBinomialObs(model, trials, states):
    nobs = states.shape[0]
    ys = np.empty((nobs, len(model.m0)))
    for t in np.arange(nobs):
        y = binom(n=trials, p=ilogit(np.dot(model.F, states[t]))).rvs()
        ys[t, :] = y
    return ys



from distutils.core import setup

setup(
    name='pssm',
    packages=['pssm'],
    version='0.0.14',
    description='Python library for state-space models',
    author='Rui Vieira',
    author_email='ruidevieira@googlemail.com',
    url='https://github.com/ruivieira/python-ssm',
    download_url='https://github.com/'
                 'ruivieira/python-ssm/archive/0.0.14.tar.gz',
    keywords=['ssm', 'statistics', 'scientific', 'state-space', 'models'],
    classifiers=[],
)
